unit Unit1;

{
  How many queens are needed to cover a chessboard?

  this program lets you find out

   - by yourself by adding queens to the board
   - by the program
     - select the number of queens
     - select SOLVE
     - press GO

  search process may be interrupted by pressing RESET
  After a solution is found, press GO to find more solutions

  DavData Software
  28-09-2015
  www.davdata.nl
}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ExtCtrls, jpeg, ComCtrls, shellApi;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    queen: TImage;
    PaintBox1: TPaintBox;
    GoBtn: TBitBtn;
    clearBtn: TBitBtn;
    tryBtn: TSpeedButton;
    msglabel: TLabel;
    solveBtn: TSpeedButton;
    UpDown1: TUpDown;
    StaticText1: TStaticText;
    Label3: TLabel;
    davImage: TImage;
    helpImage: TImage;
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure PaintBox1Paint(Sender: TObject);
    procedure tryBtnClick(Sender: TObject);
    procedure clearBtnClick(Sender: TObject);
    procedure PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure GoBtnClick(Sender: TObject);
    procedure helpImageClick(Sender: TObject);
    procedure davImageClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

type TField = record
               Q : byte; //0: empty   1:queen
               lines : byte;//1:-- 2:|  4:/  8:\
              end;
     Tpos   = record
               x,y : byte;
              end;
     TsearchResult = (srStart,srSolved,srNoSolution,srTimeout);

const mtext : array[1..13] of string =
              ('how many queens are needed to cover a chessboard?',//1
               '',                                                 //2
               'welcome at the queens puzzle',                     //3
               'solve puzzle by yourself',                         //4
               'click GO to solve',                                //5
               'field already covered',                            //6
               '',                                                 //7
               'all fields covered',                               //8
               'puzzle solved. Press GO for more solutions.',      //9
               'solving puzzle...',                                //10
               'click RESET first',                                //11
               'no solution',                                      //12
               'search interrupted');                              //13
      fieldcolors : array[0..1] of dword = ($80c0ff,$8080ff);

var Qlist : array[1..8] of Tpos; //place of queen on board
    Qcount : byte;
    board : array[1..8,1..8] of TField;
    playermode : boolean = false;

//-- solving

    maxQueen : byte;  //maximal number of queens
    searchResult : TSearchResult;
    searchbusy: boolean;
    saveI : byte;
    saveJ : byte;
    resetRequest : boolean;

procedure msg(nr : byte);
begin
 form1.msglabel.caption := mtext[nr];
end;

procedure clearpuzzle;
var i,j : byte;
begin
 Qcount := 0;
 MaxQueen := form1.UpDown1.position;
 searchResult := srStart;
 for i := 1 to 8 do
  begin
   with Qlist[i] do begin
                     x := 0; y := 0;
                    end;
   for j := 1 to 8 do
    begin
     board[i,j].Q := 0;
     board[i,j].lines := 0;
    end;
  end;
end;

procedure delay(msecs : dword);
//delay msec milliseconds
var t : dword;
begin
 t := getTickCount + msecs;
 while getTickCount < t do application.ProcessMessages;
end; 

procedure paintfield(i,j : byte);
var x,y,x1,y1,x2,y2 : word;
    n : byte;
    df : boolean;  //draw flag
begin
 x := (i-1)*60;
 y := (j-1)*60;
 with form1.PaintBox1.Canvas do
  begin
   pen.Color := $606060;
   pen.Width := 1;
   brush.Color := fieldcolors[(i xor j) and 1];
   fillrect(rect(x,y,x+60,y+60));
   if board[i,j].Q = 1 then draw(x+5,y+5,form1.queen.picture.bitmap);
   for n := 0 to 3 do
    begin
     df := false;
     case (1 shl n) and board[i,j].lines of
      1 : begin
           x1 := x;    y1 := y+30;
           x2 := x+60; y2 := y+30;
           df := true;
          end;
      2 : begin
           x1 := x+30; y1 := y;
           x2 := x+30; y2 := y+60;
           df := true;
          end;
      4 : begin
           x1 := x;    y1 := y+60;
           x2 := x+60; y2 := y;
           df := true;
          end;
      8 : begin
           x1 := x;    y1 := y;
           x2 := x+60; y2 := y+60;
           df := true;
          end;
      else begin
            x1 := 0; y1 := 0; x2 := 0; y2 := 0;
           end;
     end;//case
     if df then
      begin
       moveto(x1,y1); lineto(x2,y2);
      end; 
    end;//for n
  end;//with form1...
end;

procedure paintboard;
var i,j : byte;
begin
 for j := 1 to 8 do
  for i := 1 to 8 do paintfield(i,j);
end;

procedure fixlines(i,j : byte);
//add,delete lines at fields covered by queen (i,j)
var col,row,k,mask,n : byte;
    x,y,dx,dy : smallInt;
begin
 for n := 0 to 3 do  //directions   
  begin
   case n of
    0 : begin               //hor
         x := 1;  y := j;
         dx := 1; dy := 0;
        end;
    1 : begin               //vert
         x := i; y := 1;
         dx := 0; dy := 1;
        end;
    2 : begin
         x := i+j-8; y:= 8;
         dx := 1; dy := -1;
        end;
    3 : begin
         x := i-j+1; y := 1;
         dx := 1; dy := 1;
        end;
    else begin
          x := 0; y := 0; dx := 0; dy := 0;
         end;     
   end;//case
   mask := 1 shl n;
   for k := 0 to 7 do
    begin
     if (x >= 1) and (x <= 8) and (y >= 1) and (y <= 8) and
        ((x <> i) or (y <> j)) then
      begin
       col := x;
       row := y;
       board[col,row].lines := board[col,row].lines xor mask;
       paintfield(col,row);
      end;
     x := x + dx;
     y := y + dy;
    end;//for k
  end;//for n
end;

procedure addqueen(i,j : byte);
begin
 inc(Qcount);
 board[i,j].Q := 1;
 paintfield(i,j);
 fixlines(i,j);
 Qlist[Qcount].x := i;
 Qlist[Qcount].y := j;
end;

procedure deletequeen(i,j : byte);
begin
 board[i,j].Q := 0;
 paintfield(i,j);
 fixlines(i,j);
 Qlist[Qcount].x := 0;
 Qlist[Qcount].y := 0;
 dec(Qcount);
end;

procedure fixqueenImage;
//eliminate false colors in image
var i,j : byte;
begin
 with form1.queen.Canvas do
  for i := 0 to 49 do
   for j := 0 to 49 do
    if pixels[i,j] > 250 then pixels[i,j] := $ff;
end;

function checkSolution : boolean;
//add message if all fields covered
var i,j,n : byte;
begin
 n := 0;
 for j := 1 to 8 do
  for i:= 1 to 8 do
   with board[i,j] do
    if (Q = 1) or (lines > 0) then inc(n);
 result := n = 64;
end;

function nextfreefield(var i,j : byte) : boolean;
//supply next free (uncovered) field
var col,row : byte;
begin
 result := false;
 col := i; row := j;
 if col = 0 then row := 1;
 repeat
  inc(col);
  if col = 9 then begin
                   col := 1;
                   inc(row);
                  end;
  if row <= 8 then
    with board[col,row] do result := (q = 0) and (lines = 0);
 until result or (row = 9);
 if result then
  begin
   i := col; j := row;
  end;
end;

procedure GoSolve(var sr : TsearchResult);
//try solve puzzle for Qcount number of queens
//maxQueen has been set
var i,j : byte;
    fnext : boolean;    //solution check flag
    timeout : word;
begin
 timeout := 0;
 case sr of
  srStart      : begin
                  i := 0; j := 0;
                 end;
  srSolved     : begin
                  i := Qlist[Qcount].x;
                  j := Qlist[Qcount].y;
                 end;
  srNosolution : exit;
  srTimeout    : begin
                  i := saveI;
                  j := saveJ;
                 end;
 end;
//
 repeat
   inc(timeout);
   if timeout = 100 then
    begin
     saveI := i;
     saveJ := j;
     searchresult := srTimeout;
     exit;
    end;

   if (Qcount = MaxQueen) then deleteQueen(i,j);
   if nextFreeField(i,j) then
    begin
     fnext := true;
     addQueen(i,j);
    end
    else
     begin
      fnext := false;
      if Qcount > 0 then
       begin
        i := Qlist[Qcount].x;
        j := Qlist[Qcount].y;
        deleteQueen(i,j);
       end
       else begin
             sr := srNoSolution;
             exit;
            end;
     end;//else
 until fnext and checksolution;
 sr := srSolved;
end;

procedure procReset;
begin
 clearpuzzle;
 paintboard;
 msg(7);
 resetRequest := false;
 searchbusy := false;
end;

procedure GoProc;
begin
 repeat
  msg(10);
  GoSolve(searchresult);
  application.processmessages;
 until (searchresult <> srTimeOut) or resetRequest;
 if resetRequest then
  begin
   ProcReset;
   msg(13);
  end 
  else
   case searchresult of
    srSolved : msg(9);
    srNoSolution : msg(12);
   end;//case
end;

//-- events --

procedure TForm1.FormCreate(Sender: TObject);
begin
 resetRequest := false;
 searchbusy := false;
 playermode := true;
 label1.caption := mtext[1];
 msg(3);
 fixqueenImage;
 clearPuzzle;
 paintboard;
end;

procedure TForm1.FormPaint(Sender: TObject);
//paint edge around board
var x1,y1,x2,y2 : word;
    i : byte;
begin
 with paintbox1 do
  begin
   x1 := left-2;
   y1 := top-2;
   x2 := left + width + 1;
   y2 := top + height + 1;
  end;
 with canvas do
  begin
   pen.Width := 1;
   pen.Color := 0;
   for i := 0 to 1 do
    begin
     moveto(x1+i,y1+i);
     lineto(x1+i,y2-i);
     lineto(x2-i,y2-i);
     lineto(x2-i,y1+i);
     lineto(x1+i,y1+i);
    end;
  end;
end;

procedure TForm1.PaintBox1Paint(Sender: TObject);
begin
 paintboard;
end;

procedure TForm1.tryBtnClick(Sender: TObject);
begin
 playermode := tryBtn.Down;
 GoBtn.Enabled := not playermode;
 if playermode then
  begin
   msg(4);
   tryBtn.font.color := $000000;
   solveBtn.font.Color := $c0c0c0;
  end
  else
   begin
    msg(5);
    tryBtn.font.color := $c0c0c0;
    solveBtn.font.color := $000000;
   end;
end;

procedure TForm1.clearBtnClick(Sender: TObject);
begin
 if searchbusy then resetRequest := true
  else procReset;
end;

procedure TForm1.PaintBox1MouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
var i,j : byte;
begin
 if playermode and (shift = [ssLeft]) then
  begin
   i :=(x div 60) + 1;
   j := (y div 60) + 1;
   with board[i,j] do
    if lines = 0 then
     begin
      if board[i,j].Q = 0 then addqueen(i,j) else deletequeen(i,j);
      if checkSolution then msg(8)
       else msg(7);
     end
     else msg(6);
  end;
end;

procedure TForm1.GoBtnClick(Sender: TObject);
begin
 if searchbusy then exit;
 searchbusy := true;
 case searchresult of
  srStart : begin
             clearPuzzle;
             paintBoard;
             GoProc;
            end;
  srNoSolution : msg(11);
  srSolved : GoProc;
 end;//case
 searchbusy := false;
end;

procedure TForm1.helpImageClick(Sender: TObject);
begin
 ShellExecute(0,'open','http://www.davdata.nl/math/queens-eng.html', nil, nil, SW_SHOWNORMAL);
end;

procedure TForm1.davImageClick(Sender: TObject);
begin
 ShellExecute(0,'open','http://www.davdata.nl/math', nil, nil, SW_SHOWNORMAL);
end;

end.
